@@ -18,13 +18,13 @@
 
 #include "in_process_server.h"
 #include "thread_proxy.h"
+#include "version_specifier.h"
 #include "wlcs/display_server.h"
 #include "helpers.h"
 #include "wlcs/pointer.h"
 #include "wlcs/touch.h"
 #include "xdg_shell_v6.h"
 #include "xdg_shell_stable.h"
-#include "generated/primary-selection-unstable-v1-client.h"
 #include "generated/wayland-client.h"
 #include "generated/xdg-shell-unstable-v6-client.h"
 #include "generated/xdg-shell-client.h"
@@ -36,6 +36,7 @@
 #include <vector>
 #include <algorithm>
 #include <experimental/optional>
+#include <map>
 #include <unordered_map>
 #include <chrono>
 #include <poll.h>
@@ -55,14 +56,14 @@ public:
     }
 };
 
-wlcs::ExtensionExpectedlyNotSupported::ExtensionExpectedlyNotSupported(char const* extension, uint32_t version)
+wlcs::ExtensionExpectedlyNotSupported::ExtensionExpectedlyNotSupported(char const* extension, VersionSpecifier const& version)
     : std::runtime_error{
         std::string{"Extension: "} +
-        extension + " version " + std::to_string(version) +
+        extension + " version " + version.describe() +
         " not supported by compositor under test."}
 {
     auto const skip_reason =
-        std::string{"Missing extension: "} + extension + " v" + std::to_string(version);
+        std::string{"Missing extension: "} + extension + version.describe();
     ::testing::Test::RecordProperty("wlcs-skip-test", skip_reason);
 }
 
@@ -658,16 +659,11 @@ public:
         if (seat) wl_seat_destroy(seat);
         if (pointer) wl_pointer_destroy(pointer);
         if (touch) wl_touch_destroy(touch);
-        if (gtk_primary_selection_device_manager_)
-            gtk_primary_selection_device_manager_destroy(gtk_primary_selection_device_manager_);
-        if (primary_selection_device_manager)
-            zwp_primary_selection_device_manager_v1_destroy(primary_selection_device_manager);
-        if (data_device_manager) wl_data_device_manager_destroy(data_device_manager);
         if (xdg_shell_v6) zxdg_shell_v6_destroy(xdg_shell_v6);
         if (xdg_shell_stable) xdg_wm_base_destroy(xdg_shell_stable);
         for (auto const& output: outputs)
             wl_output_destroy(output->current.output);
-        for (auto callback: destruction_callbacks)
+        for (auto const& callback: destruction_callbacks)
             callback();
         destruction_callbacks.clear();
         wl_display_disconnect(display);
@@ -693,21 +689,6 @@ public:
         return shm;
     }
 
-    struct wl_data_device_manager* wl_data_device_manager() const
-    {
-        return data_device_manager;
-    }
-
-    struct zwp_primary_selection_device_manager_v1* zwp_primary_selection_device_manager() const
-    {
-        return primary_selection_device_manager;
-    }
-
-    struct gtk_primary_selection_device_manager* gtk_primary_selection_device_manager() const
-    {
-        return gtk_primary_selection_device_manager_;
-    }
-
     struct wl_seat* wl_seat() const
     {
         return seat;
@@ -733,7 +714,7 @@ public:
         Surface surface{client};
 
         wl_shell_surface * shell_surface = wl_shell_get_shell_surface(shell, surface);
-        run_on_destruction([shell_surface]()
+        surface.run_on_destruction([shell_surface]()
             {
                 wl_shell_surface_destroy(shell_surface);
             });
@@ -741,12 +722,8 @@ public:
 
         wl_surface_commit(surface);
 
-        surface.attach_buffer(width, height);
+        surface.attach_visible_buffer(width, height);
 
-        bool surface_rendered{false};
-        surface.add_frame_callback([&surface_rendered](auto) { surface_rendered = true; });
-        wl_surface_commit(surface);
-        dispatch_until([&surface_rendered]() { return surface_rendered; }, 10s);
         return surface;
     }
 
@@ -757,7 +734,7 @@ public:
         auto xdg_surface = std::make_shared<XdgSurfaceV6>(client, surface);
         auto xdg_toplevel = std::make_shared<XdgToplevelV6>(*xdg_surface);
 
-        run_on_destruction([xdg_surface, xdg_toplevel]() mutable
+        surface.run_on_destruction([xdg_surface, xdg_toplevel]() mutable
             {
                 xdg_surface.reset();
                 xdg_toplevel.reset();
@@ -765,12 +742,8 @@ public:
 
         wl_surface_commit(surface);
 
-        surface.attach_buffer(width, height);
+        surface.attach_visible_buffer(width, height);
 
-        bool surface_rendered{false};
-        surface.add_frame_callback([&surface_rendered](auto) { surface_rendered = true; });
-        wl_surface_commit(surface);
-        dispatch_until([&surface_rendered]() { return surface_rendered; }, 10s);
         return surface;
     }
 
@@ -781,7 +754,7 @@ public:
         auto xdg_surface = std::make_shared<XdgSurfaceStable>(client, surface);
         auto xdg_toplevel = std::make_shared<XdgToplevelStable>(*xdg_surface);
 
-        run_on_destruction([xdg_surface, xdg_toplevel]() mutable
+        surface.run_on_destruction([xdg_surface, xdg_toplevel]() mutable
             {
                 xdg_surface.reset();
                 xdg_toplevel.reset();
@@ -789,12 +762,8 @@ public:
 
         wl_surface_commit(surface);
 
-        surface.attach_buffer(width, height);
+        surface.attach_visible_buffer(width, height);
 
-        bool surface_rendered{false};
-        surface.add_frame_callback([&surface_rendered](auto) { surface_rendered = true; });
-        wl_surface_commit(surface);
-        dispatch_until([&surface_rendered]() { return surface_rendered; }, 10s);
         return surface;
     }
 
@@ -813,8 +782,12 @@ public:
         return xdg_shell_stable;
     }
 
-    wl_surface* focused_window() const
+    wl_pointer* the_pointer() const { return pointer; }
+
+    wl_surface* window_under_cursor() const
     {
+        if (pointer_events_pending())
+            BOOST_THROW_EXCEPTION(std::runtime_error("Pointer events pending"));
         if (current_pointer_location)
         {
             return current_pointer_location->surface;
@@ -824,23 +797,48 @@ public:
 
     wl_surface* touched_window() const
     {
-        if (current_touch_location)
+        if (touch_events_pending())
+            BOOST_THROW_EXCEPTION(std::runtime_error("Touch events pending"));
+        wl_surface* surface = nullptr;
+        for (auto const& touch : current_touches)
         {
-            return current_touch_location->surface;
+            if (surface && touch.second.surface != surface)
+                BOOST_THROW_EXCEPTION(std::runtime_error("Multiple surfaces have active touches"));
+            else
+                surface = touch.second.surface;
         }
-        return nullptr;
+        return surface;
     }
 
     std::pair<wl_fixed_t, wl_fixed_t> pointer_position() const
     {
+        if (pointer_events_pending())
+            BOOST_THROW_EXCEPTION(std::runtime_error("Pointer events pending"));
         return current_pointer_location.value().coordinates;
     };
 
     std::pair<wl_fixed_t, wl_fixed_t> touch_position() const
     {
-        return current_touch_location.value().coordinates;
+        if (touch_events_pending())
+            BOOST_THROW_EXCEPTION(std::runtime_error("Touch events pending"));
+        if (current_touches.empty())
+            BOOST_THROW_EXCEPTION(std::runtime_error("No touches"));
+        else if (current_touches.size() == 1)
+            return current_touches.begin()->second.coordinates;
+        else
+            BOOST_THROW_EXCEPTION(std::runtime_error("More than one touches"));
     };
 
+    bool pointer_events_pending() const
+    {
+        return !pending_buttons.empty() || pending_pointer_location;
+    }
+
+    bool touch_events_pending() const
+    {
+        return !pending_touches.empty() || !pending_up_touches.empty();
+    }
+
     void add_pointer_enter_notification(PointerEnterNotifier const& on_enter)
     {
         enter_notifiers.push_back(on_enter);
@@ -858,69 +856,59 @@ public:
         button_notifiers.push_back(on_button);
     }
 
-    void* acquire_interface(std::string const& interface, wl_interface const* to_bind, uint32_t version)
+    void* bind_if_supported(wl_interface const& to_bind, VersionSpecifier const& version) const
     {
-        wl_registry* temp_registry = wl_display_get_registry(display);
+        std::experimental::optional<bool> expected_to_be_supported{};
 
-        struct InterfaceResult
+        if (supported_extensions)
         {
-            void* bound_interface;
-            std::string const& interface;
-            wl_interface const* to_bind;
-            uint32_t version;
-        } result{nullptr, interface, to_bind, version};
-
-        wl_registry_listener const listener{
-            [](
-                void* ctx,
-                wl_registry* registry,
-                uint32_t id,
-                char const* interface,
-                uint32_t version)
-            {
-                auto request = static_cast<InterfaceResult*>(ctx);
-
-                if ((request->interface == interface) &&
-                    version >= request->version)
-                {
-                    request->bound_interface =
-                        wl_registry_bind(registry, id, request->to_bind, version);
-                }
-            },
-            [](auto, auto, auto) {}
-        };
-
-        wl_registry_add_listener(
-            temp_registry,
-            &listener,
-            &result);
-
-        wl_display_roundtrip(display);
-
-        wl_registry_destroy(temp_registry);
-
-        if (result.bound_interface == nullptr)
-        {
-            if (
-                supported_extensions &&
-                (supported_extensions->count(interface.c_str()) == 0 ||
-                 supported_extensions->at(interface.c_str()) < version))
-            {
-                BOOST_THROW_EXCEPTION((ExtensionExpectedlyNotSupported{interface.c_str(), version}));
-            }
-            else
-            {
-                BOOST_THROW_EXCEPTION((std::runtime_error{"Failed to acquire interface"}));
-            }
+            expected_to_be_supported =
+                supported_extensions->count(to_bind.name) &&
+                version.select_version(supported_extensions->at(to_bind.name));
         }
 
-        return result.bound_interface;
+        /* TODO: Mark tests using globals which exist, but are listed as unsupported as
+         * may-fail.
+         * This should help incrementally implementing a protocol, while getting test
+         * feedback.
+         */
+        auto const global = globals.find(to_bind.name);
+        if (global != globals.end())
+        {
+            auto const version_to_bind = version.select_version(global->second.version);
+            if (version_to_bind)
+            {
+                auto global_proxy = wl_registry_bind(registry, global->second.id, &to_bind, version_to_bind.value());
+                if (!global_proxy)
+                {
+                    throw_wayland_error(display);
+                }
+                return global_proxy;
+            }
+            else if (!expected_to_be_supported.value_or(true))
+            {
+                // We didn't expect to find the needed version of this extension anyway…
+                BOOST_THROW_EXCEPTION((ExtensionExpectedlyNotSupported{to_bind.name, version}));
+            }
+
+            BOOST_THROW_EXCEPTION((std::runtime_error{
+                "Failed to bind to " + std::string{to_bind.name} + " version " + version.describe()}));
+        }
+        else if (!expected_to_be_supported.value_or(true))
+        {
+            // We didn't expect to find this extension anyway…
+            BOOST_THROW_EXCEPTION((ExtensionExpectedlyNotSupported{to_bind.name, version}));
+        }
+        BOOST_THROW_EXCEPTION((std::runtime_error{
+            "Failed to bind to " + std::string{to_bind.name} + " version " + version.describe()}));
     }
 
     void dispatch_until(
         std::function<bool()> const& predicate, std::chrono::seconds timeout)
     {
-        auto const end_time = std::chrono::steady_clock::now() + timeout;
+        using namespace std::chrono;
+
+        auto const end_time = steady_clock::now() + timeout;
         while (!predicate())
         {
             while (wl_display_prepare_read(display) != 0)
@@ -930,14 +918,22 @@ public:
             }
             wl_display_flush(display);
 
-            auto const maximum_wait = end_time - std::chrono::steady_clock::now();
-            if (maximum_wait.count() < 0)
+            auto const time_left = end_time - steady_clock::now();
+            if (time_left.count() < 0)
             {
                 BOOST_THROW_EXCEPTION((Timeout{"Timeout waiting for condition"}));
             }
 
-            auto const maximum_wait_ms =
-                std::chrono::duration_cast<std::chrono::milliseconds>(maximum_wait);
+            /*
+             * TODO: We really want std::chrono::duration::ceil() here, but that's C++17
+             */
+            /*
+             * We want to wait *at least* as long as time_left. duration_cast<milliseconds>
+             * will perform integer division, so any fractional milliseconds will get dropped.
+             *
+             * Adding 1ms will ensure we wait until *after* we're meant to timeout.
+             */
+            auto const maximum_wait_ms = duration_cast<milliseconds>(time_left) + 1ms;
             pollfd fd{
                 wl_display_get_fd(display),
                 POLLIN | POLLERR,
@@ -984,6 +980,7 @@ public:
     {
         OutputState current;
         OutputState pending;
+        std::vector<std::function<void()>> done_notifiers;
 
         Output(struct wl_output* output)
             : current{output},
@@ -1031,6 +1028,9 @@ public:
                 me->current.scale = me->pending.scale;
 
             me->pending = OutputState{me->current.output};
+
+            for (auto const& notifier : me->done_notifiers)
+                notifier();
         }
 
         static void scale_thunk(void* ctx, struct wl_output */*wl_output*/, int32_t factor)
@@ -1060,23 +1060,15 @@ private:
     {
         auto me = static_cast<Impl*>(ctx);
 
-        me->current_pointer_location = SurfaceLocation {
-            surface,
-            std::make_pair(x,y)
-        };
+        if (me->current_pointer_location && !me->pending_pointer_leave)
+            FAIL()
+                << "Pointer tried to enter surface " << surface
+                << " without first leaving surface " << me->current_pointer_location.value().surface;
 
-        std::vector<decltype(enter_notifiers)::const_iterator> to_remove;
-        for (auto notifier = me->enter_notifiers.begin(); notifier != me->enter_notifiers.end(); ++notifier)
-        {
-            if (!(*notifier)(surface, x, y))
-            {
-                to_remove.push_back(notifier);
-            }
-        }
-        for (auto removed : to_remove)
-        {
-            me->enter_notifiers.erase(removed);
-        }
+        me->pending_pointer_location = SurfaceLocation{
+            surface,
+            std::make_pair(x, y)
+        };
     }
 
     static void pointer_leave(
@@ -1087,20 +1079,17 @@ private:
     {
         auto me = static_cast<Impl*>(ctx);
 
-        me->current_pointer_location = {};
+        if (!me->current_pointer_location)
+            FAIL() << "Got wl_pointer.leave when the pointer was not on a surface";
 
-        std::vector<decltype(leave_notifiers)::const_iterator> to_remove;
-        for (auto notifier = me->leave_notifiers.begin(); notifier != me->leave_notifiers.end(); ++notifier)
-        {
-            if (!(*notifier)(surface))
-            {
-                to_remove.push_back(notifier);
-            }
-        }
-        for (auto removed : to_remove)
-        {
-            me->leave_notifiers.erase(removed);
-        }
+        // the surface should never be null along the wire, but may come out as null if it's been destroyed
+        if (surface != nullptr && surface != me->current_pointer_location.value().surface)
+            FAIL()
+                << "Got wl_pointer.leave with surface " << surface
+                << " instead of " << me->current_pointer_location.value().surface;
+
+        me->pending_pointer_location = std::experimental::nullopt;
+        me->pending_pointer_leave = true;
     }
 
     static void pointer_motion(
@@ -1112,20 +1101,12 @@ private:
     {
         auto me = static_cast<Impl*>(ctx);
 
-        me->current_pointer_location.value().coordinates = std::make_pair(x, y);
+        if (!me->current_pointer_location && !me->pending_pointer_location)
+            FAIL() << "Got wl_pointer.motion when the pointer was not on a surface";
 
-        std::vector<decltype(motion_notifiers)::const_iterator> to_remove;
-        for (auto notifier = me->motion_notifiers.begin(); notifier != me->motion_notifiers.end(); ++notifier)
-        {
-            if (!(*notifier)(x, y))
-            {
-                to_remove.push_back(notifier);
-            }
-        }
-        for (auto removed : to_remove)
-        {
-            me->motion_notifiers.erase(removed);
-        }
+        if (!me->pending_pointer_location)
+            me->pending_pointer_location = me->current_pointer_location;
+        me->pending_pointer_location.value().coordinates = std::make_pair(x, y);
     }
 
     static void pointer_button(
@@ -1138,22 +1119,116 @@ private:
     {
         auto me = static_cast<Impl*>(ctx);
 
-        std::vector<decltype(button_notifiers)::const_iterator> to_remove;
-        for (auto notifier = me->button_notifiers.begin(); notifier != me->button_notifiers.end(); ++notifier)
+        me->pending_buttons[button] = std::make_pair(serial, state == WL_POINTER_BUTTON_STATE_PRESSED);
+    }
+
+    static void pointer_frame(void* ctx, wl_pointer* /*pointer*/)
+    {
+        auto me = static_cast<Impl*>(ctx);
+
+        if (me->pending_pointer_leave)
         {
-            if (!(*notifier)(serial, button, state == WL_POINTER_BUTTON_STATE_PRESSED))
+            if (!me->current_pointer_location)
+                FAIL() << "Pointer tried to leave when it was not on a surface";
+
+            wl_surface* old_surface = me->current_pointer_location.value().surface;
+            me->current_pointer_location = std::experimental::nullopt;
+            me->pending_pointer_leave = false;
+
+            me->notify_of_pointer_leave(old_surface);
+        }
+
+        if (me->pending_pointer_location)
+        {
+            auto const old_pointer_location = me->current_pointer_location;
+            me->current_pointer_location = me->pending_pointer_location;
+            me->pending_pointer_location = std::experimental::nullopt;
+
+            if (!old_pointer_location)
+            {
+                me->notify_of_pointer_enter(
+                    me->current_pointer_location.value().surface,
+                    me->current_pointer_location.value().coordinates);
+            }
+            else
+            {
+                me->notify_of_pointer_motion(me->current_pointer_location.value().coordinates);
+            }
+        }
+
+        if (!me->pending_buttons.empty())
+        {
+            me->notify_of_pointer_buttons(me->pending_buttons);
+            me->pending_buttons.clear();
+        }
+    }
+
+    void notify_of_pointer_enter(wl_surface* surface, std::pair<int, int> position)
+    {
+        std::vector<decltype(enter_notifiers)::const_iterator> to_remove;
+        for (auto notifier = enter_notifiers.begin(); notifier != enter_notifiers.end(); ++notifier)
+        {
+            if (!(*notifier)(surface, position.first, position.second))
             {
                 to_remove.push_back(notifier);
             }
         }
         for (auto removed : to_remove)
         {
-            me->button_notifiers.erase(removed);
+            enter_notifiers.erase(removed);
         }
     }
 
-    static void pointer_frame(void* /*ctx*/, wl_pointer* /*pointer*/)
+    void notify_of_pointer_leave(wl_surface* surface)
     {
+        std::vector<decltype(leave_notifiers)::const_iterator> to_remove;
+        for (auto notifier = leave_notifiers.begin(); notifier != leave_notifiers.end(); ++notifier)
+        {
+            if (!(*notifier)(surface))
+            {
+                to_remove.push_back(notifier);
+            }
+        }
+        for (auto removed : to_remove)
+        {
+            leave_notifiers.erase(removed);
+        }
+    }
+
+    void notify_of_pointer_motion(std::pair<int, int> position)
+    {
+
+        std::vector<decltype(motion_notifiers)::const_iterator> to_remove;
+        for (auto notifier = motion_notifiers.begin(); notifier != motion_notifiers.end(); ++notifier)
+        {
+            if (!(*notifier)(position.first, position.second))
+            {
+                to_remove.push_back(notifier);
+            }
+        }
+        for (auto removed : to_remove)
+        {
+            motion_notifiers.erase(removed);
+        }
+    }
+
+    void notify_of_pointer_buttons(std::map<uint32_t, std::pair<uint32_t, bool>> const& buttons)
+    {
+        for (auto const& button : buttons)
+        {
+            std::vector<decltype(button_notifiers)::const_iterator> to_remove;
+            for (auto notifier = button_notifiers.begin(); notifier != button_notifiers.end(); ++notifier)
+            {
+                if (!(*notifier)(button.second.first, button.first, button.second.second))
+                {
+                    to_remove.push_back(notifier);
+                }
+            }
+            for (auto removed : to_remove)
+            {
+                button_notifiers.erase(removed);
+            }
+        }
     }
 
     static constexpr wl_pointer_listener pointer_listener = {
@@ -1174,13 +1249,17 @@ private:
         uint32_t /*serial*/,
         uint32_t /*time*/,
         wl_surface* surface,
-        int32_t /*id*/,
+        int32_t id,
         wl_fixed_t x,
         wl_fixed_t y)
     {
         auto me = static_cast<Impl*>(ctx);
 
-        me->current_touch_location = SurfaceLocation {
+        auto touch = me->current_touches.find(id);
+        if (touch != me->current_touches.end())
+            FAIL() << "Got wl_touch.down with ID " << id << " which is already down";
+
+        me->pending_touches[id] = SurfaceLocation {
             surface,
             std::make_pair(x,y)
         };
@@ -1191,29 +1270,53 @@ private:
         wl_touch* /*wl_touch*/,
         uint32_t /*serial*/,
         uint32_t /*time*/,
-        int32_t /*id*/)
+        int32_t id)
     {
         auto me = static_cast<Impl*>(ctx);
 
-        me->current_touch_location = std::experimental::nullopt;
+        auto touch = me->current_touches.find(id);
+        if (touch == me->current_touches.end())
+            FAIL() << "Got wl_touch.up with unknown ID " << id;
+
+        me->pending_up_touches.insert(id);
     }
 
 	static void touch_motion(
         void* ctx,
         wl_touch* /*wl_touch*/,
         uint32_t /*time*/,
-        int32_t /*id*/,
+        int32_t id,
         wl_fixed_t x,
         wl_fixed_t y)
     {
         auto me = static_cast<Impl*>(ctx);
 
-        if (me->current_touch_location)
-            me->current_touch_location.value().coordinates = std::make_pair(x,y);
+        auto touch = me->current_touches.find(id);
+        if (touch == me->current_touches.end())
+            FAIL() << "Got wl_touch.up with unknown ID " << id;
+
+        me->pending_touches[id] = SurfaceLocation {
+            touch->second.surface,
+            std::make_pair(x,y)
+        };
     }
 
-    static void touch_frame(void* /*ctx*/, wl_touch* /*touch*/)
+    static void touch_frame(void* ctx, wl_touch* /*touch*/)
     {
+        auto me = static_cast<Impl*>(ctx);
+
+        for (auto const& id : me->pending_up_touches)
+        {
+            me->current_touches.erase(id);
+        }
+
+        for (auto const& touch : me->pending_touches)
+        {
+            me->current_touches[touch.first] = touch.second;
+        }
+
+        me->pending_up_touches.clear();
+        me->pending_touches.clear();
     }
 
     static constexpr wl_touch_listener touch_listener = {
@@ -1268,6 +1371,8 @@ private:
         using namespace std::literals::string_literals;
 
         auto me = static_cast<Impl*>(ctx);
+        me->global_type_names[id] = interface;
+        me->globals[interface] = Global{id, version};
 
         if ("wl_shm"s == interface)
         {
@@ -1298,21 +1403,6 @@ private:
             // Ensure we receive the initial seat events.
             me->server_roundtrip();
         }
-        else if ("wl_data_device_manager"s == interface)
-        {
-            me->data_device_manager = static_cast<struct wl_data_device_manager*>(
-                wl_registry_bind(registry, id, &wl_data_device_manager_interface, version));
-        }
-        else if ("zwp_primary_selection_device_manager_v1"s == interface)
-        {
-            me->primary_selection_device_manager = static_cast<struct zwp_primary_selection_device_manager_v1*>(
-                wl_registry_bind(registry, id, &zwp_primary_selection_device_manager_v1_interface, version));
-        }
-        else if ("gtk_primary_selection_device_manager"s == interface)
-        {
-            me->gtk_primary_selection_device_manager_ = static_cast<struct gtk_primary_selection_device_manager*>(
-                    wl_registry_bind(registry, id, &gtk_primary_selection_device_manager_interface, version));
-        }
         else if ("wl_output"s == interface)
         {
             auto wl_output = static_cast<struct wl_output*>(
@@ -1336,9 +1426,15 @@ private:
         }
     }
 
-    static void global_removed(void*, wl_registry*, uint32_t)
+    static void global_removed(void* ctx, wl_registry*, uint32_t id)
     {
-        // TODO: Remove our globals
+        auto me = static_cast<Impl*>(ctx);
+        auto const name = me->global_type_names.find(id);
+        if (name != me->global_type_names.end())
+        {
+            me->globals.erase(name->second);
+            me->global_type_names.erase(name);
+        }
     }
 
     constexpr static wl_registry_listener registry_listener = {
@@ -1357,12 +1453,17 @@ private:
     struct wl_seat* seat = nullptr;
     struct wl_pointer* pointer = nullptr;
     struct wl_touch* touch = nullptr;
-    struct wl_data_device_manager* data_device_manager = nullptr;
     struct zxdg_shell_v6* xdg_shell_v6 = nullptr;
     std::vector<std::function<void()>> destruction_callbacks;
     struct xdg_wm_base* xdg_shell_stable = nullptr;
-    struct zwp_primary_selection_device_manager_v1* primary_selection_device_manager = nullptr;
-    struct gtk_primary_selection_device_manager* gtk_primary_selection_device_manager_ = nullptr;
+
+    struct Global
+    {
+        uint32_t id;
+        uint32_t version;
+    };
+    std::map<std::string, Global> globals;
+    std::map<uint32_t, std::string> global_type_names;
 
     struct SurfaceLocation
     {
@@ -1370,7 +1471,12 @@ private:
         std::pair<wl_fixed_t, wl_fixed_t> coordinates;
     };
     std::experimental::optional<SurfaceLocation> current_pointer_location;
-    std::experimental::optional<SurfaceLocation> current_touch_location;
+    std::experimental::optional<SurfaceLocation> pending_pointer_location;
+    bool pending_pointer_leave{false};
+    std::map<uint32_t, std::pair<uint32_t, bool>> pending_buttons; ///< Maps button id to the serial and if it's down
+    std::map<int, SurfaceLocation> current_touches; ///< Touches that have gotten a frame event
+    std::map<int, SurfaceLocation> pending_touches; ///< Touches that have gotten down or motion events without a frame
+    std::set<int> pending_up_touches; ///< Touches that have gotten up events without a frame
 
     std::vector<PointerEnterNotifier> enter_notifiers;
     std::vector<PointerLeaveNotifier> leave_notifiers;
@@ -1411,21 +1517,6 @@ wl_shm* wlcs::Client::shm() const
     return impl->wl_shm();
 }
 
-struct wl_data_device_manager* wlcs::Client::data_device_manager() const
-{
-    return impl->wl_data_device_manager();
-}
-
-struct zwp_primary_selection_device_manager_v1* wlcs::Client::primary_selection_device_manager() const
-{
-    return impl->zwp_primary_selection_device_manager();
-}
-
-struct gtk_primary_selection_device_manager* wlcs::Client::gtk_primary_selection_device_manager() const
-{
-    return impl->gtk_primary_selection_device_manager();
-}
-
 wl_seat* wlcs::Client::seat() const
 {
     return impl->wl_seat();
@@ -1469,11 +1560,19 @@ size_t wlcs::Client::output_count() const
 wlcs::OutputState wlcs::Client::output_state(size_t index) const
 {
     if (index > output_count())
-        throw std::runtime_error("Invalid output index");
+        throw std::out_of_range("Invalid output index");
 
     return impl->outputs[index]->current;
 }
 
+void wlcs::Client::add_output_done_notifier(size_t index, std::function<void()> const& notifier)
+{
+    if (index > output_count())
+        throw std::runtime_error("Invalid output index");
+
+    impl->outputs[index]->done_notifiers.push_back(notifier);
+}
+
 wl_shell* wlcs::Client::shell() const
 {
     return impl->the_shell();
@@ -1489,9 +1588,9 @@ xdg_wm_base* wlcs::Client::xdg_shell_stable() const
     return impl->the_xdg_shell_stable();
 }
 
-wl_surface* wlcs::Client::focused_window() const
+wl_surface* wlcs::Client::window_under_cursor() const
 {
-    return impl->focused_window();
+    return impl->window_under_cursor();
 }
 
 wl_surface* wlcs::Client::touched_window() const
@@ -1539,12 +1638,14 @@ void wlcs::Client::roundtrip()
     impl->server_roundtrip();
 }
 
-void* wlcs::Client::acquire_interface(
-    std::string const& name,
-    wl_interface const* interface,
-    uint32_t version)
+void* wlcs::Client::bind_if_supported(wl_interface const& interface, VersionSpecifier const& version) const
 {
-    return impl->acquire_interface(name, interface, version);
+    return impl->bind_if_supported(interface, version);
+}
+
+wl_pointer* wlcs::Client::the_pointer() const
+{
+    return impl->the_pointer();
 }
 
 class wlcs::Surface::Impl
@@ -1574,6 +1675,10 @@ public:
             }
         }
 
+        for (auto const& callback: destruction_callbacks)
+            callback();
+        destruction_callbacks.clear();
+
         wl_surface_destroy(surface_);
     }
 
@@ -1600,6 +1705,20 @@ public:
         wl_callback_add_listener(callback, &frame_listener, holder.release());
     }
 
+    void attach_visible_buffer(int width, int height)
+    {
+        attach_buffer(width, height);
+        auto surface_rendered = std::make_shared<bool>(false);
+        add_frame_callback([surface_rendered](auto) { *surface_rendered = true; });
+        wl_surface_commit(surface_);
+        owner_.dispatch_until([surface_rendered]() { return *surface_rendered; });
+    }
+
+    void run_on_destruction(std::function<void()> callback)
+    {
+        destruction_callbacks.push_back(callback);
+    }
+
     Client& owner() const
     {
         return owner_;
@@ -1633,6 +1752,7 @@ private:
 
     struct wl_surface* const surface_;
     Client& owner_;
+    std::vector<std::function<void()>> destruction_callbacks;
 };
 
 std::vector<std::pair<wlcs::Surface::Impl const*, wl_callback*>> wlcs::Surface::Impl::pending_callbacks;
@@ -1663,6 +1783,16 @@ void wlcs::Surface::add_frame_callback(std::function<void(int)> const& on_frame)
     impl->add_frame_callback(on_frame);
 }
 
+void wlcs::Surface::attach_visible_buffer(int width, int height)
+{
+    impl->attach_visible_buffer(width, height);
+}
+
+void wlcs::Surface::run_on_destruction(std::function<void()> callback)
+{
+    impl->run_on_destruction(callback);
+}
+
 wlcs::Client& wlcs::Surface::owner() const
 {
     return impl->owner();
@@ -1808,8 +1938,3 @@ void wlcs::ShmBuffer::add_release_listener(std::function<bool()> const &on_relea
 {
     impl->add_release_listener(on_release);
 }
-
-wlcs::CheckInterfaceExpected::CheckInterfaceExpected(Server& server, wl_interface const& interface) : Client{server}
-{
-    acquire_interface(interface.name, &interface, interface.version);
-}
