@@ -29,6 +29,7 @@
 #include "in_process_server.h"
 #include "xdg_shell_stable.h"
 #include "xdg_shell_v6.h"
+#include "layer_shell_v1.h"
 
 #include <gmock/gmock.h>
 
@@ -261,7 +262,8 @@ public:
 
 
         popup_xdg_surface.emplace(client, popup_surface.value());
-        popup.emplace(popup_xdg_surface.value(), xdg_shell_surface, positioner);
+
+        popup.emplace(popup_xdg_surface.value(), &xdg_shell_surface, positioner);
 
         popup_xdg_surface.value().add_configure_notification([&](uint32_t serial)
             {
@@ -390,6 +392,89 @@ public:
     int popup_surface_configure_count{0};
 };
 
+class LayerV1PopupManager : public XdgPopupManagerBase
+{
+public:
+    LayerV1PopupManager(wlcs::InProcessServer* const in_process_server)
+        : XdgPopupManagerBase{in_process_server},
+          layer_surface{client, surface}
+    {
+        zwlr_layer_surface_v1_set_size(layer_surface, window_width, window_height);
+        wait_for_frame_to_render();
+    }
+
+    void dispatch_until_popup_configure() override
+    {
+        client.dispatch_until(
+            [prev_count = popup_surface_configure_count, &current_count = popup_surface_configure_count]()
+            {
+                return current_count > prev_count;
+            });
+    }
+
+    void setup_popup(PositionerParams const& param) override
+    {
+        wlcs::XdgPositionerStable positioner{client};
+
+        // size must always be set
+        xdg_positioner_set_size(positioner, param.popup_size.first, param.popup_size.second);
+
+        // anchor rect must always be set
+        xdg_positioner_set_anchor_rect(
+            positioner,
+            param.anchor_rect.first.first,
+            param.anchor_rect.first.second,
+            param.anchor_rect.second.first,
+            param.anchor_rect.second.second);
+
+        if (param.anchor_stable)
+            xdg_positioner_set_anchor(positioner,  param.anchor_stable.value());
+
+        if (param.gravity_stable)
+            xdg_positioner_set_gravity(positioner, param.gravity_stable.value());
+
+        if (param.constraint_adjustment_stable)
+            xdg_positioner_set_constraint_adjustment(positioner, param.constraint_adjustment_stable.value());
+
+        if (param.offset)
+            xdg_positioner_set_offset(positioner, param.offset.value().first, param.offset.value().second);
+
+
+        popup_xdg_surface.emplace(client, popup_surface.value());
+        popup.emplace(popup_xdg_surface.value(), std::experimental::nullopt, positioner);
+        zwlr_layer_surface_v1_get_popup(layer_surface, popup.value());
+
+        popup_xdg_surface.value().add_configure_notification([&](uint32_t serial)
+            {
+                xdg_surface_ack_configure(popup_xdg_surface.value(), serial);
+                popup_surface_configure_count++;
+            });
+
+        popup.value().add_configure_notification([this](int32_t x, int32_t y, int32_t width, int32_t height)
+            {
+                state = State{x, y, width, height};
+            });
+    }
+
+    void clear_popup() override
+    {
+        popup = std::experimental::nullopt;
+        popup_xdg_surface = std::experimental::nullopt;
+    }
+
+    auto popup_position() const -> std::experimental::optional<std::pair<int, int>> override
+    {
+        return std::make_pair(state.value().x, state.value().y);
+    }
+
+    wlcs::LayerSurfaceV1 layer_surface;
+
+    std::experimental::optional<wlcs::XdgSurfaceStable> popup_xdg_surface;
+    std::experimental::optional<wlcs::XdgPopupStable> popup;
+
+    int popup_surface_configure_count{0};
+};
+
 }
 
 class XdgPopupPositionerTest:
@@ -430,14 +515,30 @@ TEST_P(XdgPopupPositionerTest, xdg_shell_unstable_v6_popup_placed_correctly)
         Eq(std::experimental::make_optional(param.expected_positon))) << "popup placed in incorrect position";
 }
 
-INSTANTIATE_TEST_CASE_P(
+TEST_P(XdgPopupPositionerTest, layer_shell_popup_placed_correctly)
+{
+    auto manager = std::make_unique<LayerV1PopupManager>(this);
+    auto const& param = GetParam();
+
+    manager->map_popup(param.positioner);
+
+    ASSERT_THAT(
+        manager->popup_position(),
+        Ne(std::experimental::nullopt)) << "popup configure event not sent";
+
+    ASSERT_THAT(
+        manager->popup_position(),
+        Eq(std::experimental::make_optional(param.expected_positon))) << "popup placed in incorrect position";
+}
+
+INSTANTIATE_TEST_SUITE_P(
     Default,
     XdgPopupPositionerTest,
     testing::Values(
         PositionerTestParams{"default values", (window_width - popup_width) / 2, (window_height - popup_height) / 2, PositionerParams()}
     ));
 
-INSTANTIATE_TEST_CASE_P(
+INSTANTIATE_TEST_SUITE_P(
     Anchor,
     XdgPopupPositionerTest,
     testing::Values(
@@ -466,7 +567,7 @@ INSTANTIATE_TEST_CASE_P(
             PositionerParams().with_anchor(XDG_POSITIONER_ANCHOR_BOTTOM_RIGHT)}
     ));
 
-INSTANTIATE_TEST_CASE_P(
+INSTANTIATE_TEST_SUITE_P(
     Gravity,
     XdgPopupPositionerTest,
     testing::Values(
@@ -498,7 +599,7 @@ INSTANTIATE_TEST_CASE_P(
             PositionerParams().with_gravity(XDG_POSITIONER_GRAVITY_BOTTOM_RIGHT)}
     ));
 
-INSTANTIATE_TEST_CASE_P(
+INSTANTIATE_TEST_SUITE_P(
     AnchorRect,
     XdgPopupPositionerTest,
     testing::Values(
@@ -542,7 +643,7 @@ TEST_P(XdgPopupTest, pointer_focus_goes_to_popup)
     pointer.move_to(manager->window_x + 1, manager->window_y + 1);
     manager->client.roundtrip();
 
-    EXPECT_THAT(manager->client.focused_window(), Eq((wl_surface*)manager->surface));
+    EXPECT_THAT(manager->client.window_under_cursor(), Eq((wl_surface*)manager->surface));
 
     auto positioner = PositionerParams{}
         .with_size(30, 30)
@@ -554,7 +655,7 @@ TEST_P(XdgPopupTest, pointer_focus_goes_to_popup)
     pointer.move_to(manager->window_x + 2, manager->window_y + 1);
     manager->client.roundtrip();
 
-    EXPECT_THAT(manager->client.focused_window(), Eq((wl_surface*)manager->popup_surface.value()));
+    EXPECT_THAT(manager->client.window_under_cursor(), Eq((wl_surface*)manager->popup_surface.value()));
 }
 
 TEST_P(XdgPopupTest, popup_gives_up_pointer_focus_when_gone)
@@ -573,14 +674,14 @@ TEST_P(XdgPopupTest, popup_gives_up_pointer_focus_when_gone)
     pointer.move_to(manager->window_x + 2, manager->window_y + 1);
     manager->client.roundtrip();
 
-    EXPECT_THAT(manager->client.focused_window(), Eq((wl_surface*)manager->popup_surface.value()));
+    EXPECT_THAT(manager->client.window_under_cursor(), Eq((wl_surface*)manager->popup_surface.value()));
 
     manager->unmap_popup();
     manager->client.roundtrip();
     pointer.move_to(manager->window_x + 3, manager->window_y + 1);
     manager->client.roundtrip();
 
-    EXPECT_THAT(manager->client.focused_window(), Eq((wl_surface*)manager->surface));
+    EXPECT_THAT(manager->client.window_under_cursor(), Eq((wl_surface*)manager->surface));
 }
 
 TEST_F(XdgPopupTest, zero_size_anchor_rect_stable)
@@ -619,7 +720,7 @@ TEST_P(XdgPopupTest, popup_configure_is_valid)
     EXPECT_THAT(manager->state.value().height, Gt(0));
 }
 
-INSTANTIATE_TEST_CASE_P(
+INSTANTIATE_TEST_SUITE_P(
     XdgPopupStable,
     XdgPopupTest,
     testing::Values(XdgPopupTestParam{
@@ -628,7 +729,7 @@ INSTANTIATE_TEST_CASE_P(
             return std::make_unique<XdgPopupStableManager>(server);
         }}));
 
-INSTANTIATE_TEST_CASE_P(
+INSTANTIATE_TEST_SUITE_P(
     XdgPopupUnstableV6,
     XdgPopupTest,
     testing::Values(XdgPopupTestParam{
@@ -637,6 +738,15 @@ INSTANTIATE_TEST_CASE_P(
             return std::make_unique<XdgPopupV6Manager>(server);
         }}));
 
+INSTANTIATE_TEST_SUITE_P(
+    LayerShellPopup,
+    XdgPopupTest,
+    testing::Values(XdgPopupTestParam{
+        [](wlcs::InProcessServer* const server)
+        {
+            return std::make_unique<LayerV1PopupManager>(server);
+        }}));
+
 // TODO: test that positioner is always overlapping or adjacent to parent
 // TODO: test that positioner is copied immediately after use
 // TODO: test that error is raised when incomplete positioner is used (positioner without size and anchor rect set)
